/*
 * ** header v3.0
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2018 Research Group Biomedical Physics,
 * Max-Planck-Institute for Dynamics and Self-Organization Göttingen
 * Copyright (C) 2019 IndiScale GmbH
 * Copyright (C) 2019 Timm Fitschen (t.fitschen@indiscale.com)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * ** end header
 */
package org.caosdb.server.database.backend.transaction;

import java.util.ArrayList;
import org.apache.commons.jcs.access.behavior.ICacheAccess;
import org.caosdb.server.caching.Cache;
import org.caosdb.server.database.CacheableBackendTransaction;
import org.caosdb.server.database.backend.interfaces.RuleLoaderImpl;
import org.caosdb.server.database.exceptions.TransactionException;
import org.caosdb.server.database.proto.Rule;
import org.caosdb.server.entity.EntityInterface;
import org.caosdb.server.entity.container.TransactionContainer;
import org.caosdb.server.entity.wrapper.Property;
import org.caosdb.server.jobs.Job;
import org.caosdb.server.transaction.Transaction;

/**
 * Load transaction rules (e.g. "Referenced entities must always exist") and their configuration
 * from the back-end.
 *
 * @author Timm Fitschen (t.fitschen@indiscale.com)
 */
public class RuleLoader extends CacheableBackendTransaction<String, ArrayList<Rule>> {

  private static final ICacheAccess<String, ArrayList<Rule>> cache =
      Cache.getCache("BACKEND_JobRules");
  private final Transaction<? extends TransactionContainer> transaction;
  private final EntityInterface e;
  private final Integer entity;
  private final Integer domain;
  private ArrayList<Job> jobs;
  private String transactionType;

  public RuleLoader(
      final Integer domain,
      final Integer entity,
      final EntityInterface e,
      final Transaction<? extends TransactionContainer> transaction) {
    super(cache);
    this.domain = domain;
    this.entity = entity;
    this.e = e;
    if (e instanceof Property) {
      this.transactionType = getTransactionType(((Property) e).getDomainEntity());
    } else {
      this.transactionType = getTransactionType(e);
    }
    this.transaction = transaction;
  }

  @Override
  protected String getKey() {
    return "<" + this.domain + "," + this.entity + "," + this.transactionType + ">";
  }

  public ArrayList<Job> getJobs() {
    return this.jobs;
  }

  @Override
  public ArrayList<Rule> executeNoCache() throws TransactionException {
    final RuleLoaderImpl t = getImplementation(RuleLoaderImpl.class);
    return t.executeNoCache(this.domain, this.entity, this.transactionType);
  }

  @Override
  protected void process(final ArrayList<Rule> t) throws TransactionException {
    this.jobs = new ArrayList<Job>();
    for (final Rule r : t) {
      this.jobs.add(Job.getJob(r.job, r.mode, this.e, this.transaction));
    }
  }
}
